#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : sky.li (13751020500@163.com)
# @Desc    : 数据包解析工具类

import asyncio

# 字节流解析
class DataStream:
    def __init__(self):
        self._datas = []
        self._reqCount = -1
        self._recvFuture = None  # 接收字节future
        self._toMatch = []  # 匹配头
        self._recvTillFuture = None  # 接收匹配字节 future
        self._recvLoop = asyncio.get_event_loop()

    # 字节流收到匹配 buff
    async def recv(self, n):
        self._reqCount = n
        self._recvFuture = asyncio.Future()
        self._recvLoop.call_soon_threadsafe(self._tryRecv)
        return await self._recvFuture

    # 字节流收到匹配 buff

    def isEmpty(self):
        return len(self._datas) == 0

    # 字节流收到匹配 buff
    async def recvTill(self, buff):
        self._toMatch = []
        for byte in buff:
            self._toMatch.append(byte)
        self._recvTillFuture = asyncio.Future()
        self._recvLoop.call_soon_threadsafe(self._tryRecvTill)
        await self._recvTillFuture

    def _tryRecvTill(self):
        if self._recvTillFuture == None:
            return
        while True:  # lable tryRecv
            noMatch = False
            for i in range(len(self._toMatch)):
                # no more byte now
                if i == len(self._datas):
                    return
                if self._toMatch[i] != self._datas[i]:
                    self._datas.pop(0)
                    noMatch = True
                    break
            if not noMatch:
                break

        res = self._recvTillFuture
        self._recvTillFuture = None
        self._datas = self._datas[len(self._toMatch):]
        self._toMatch = []
        res.set_result(None)

    def _tryRecv(self):
        if (self._recvFuture == None):
            return
        if (len(self._datas) >= self._reqCount):
            rest = self._datas[self._reqCount:]
            nbytes = self._datas[:self._reqCount]
            self._datas = rest
            self._reqCount = -1

            res = self._recvFuture
            self._recvFuture = None
            res.set_result(nbytes)
    
    # 填入数据
    def feedBytes(self, data):
        for byte in data:
            self._datas.append(byte)
        self._tryRecv()
        self._tryRecvTill()


# def __test():
#     ds = DataStream()
#     head = [b for b in bytes("_8weurswea_", 'utf-8')]

#     async def t1():
#         data = await ds.recv(2)
#         assert (data == [1, 2])
#         data = await ds.recv(3)
#         assert (data == [3, 4, 5])
#         await ds.recvTill(head)
#         assert (ds._datas == [11, 23])

#     async def t2():
#         await asyncio.sleep(1)
#         ds.feedBytes([1])
#         ds.feedBytes([2])
#         ds.feedBytes([3, 4])
#         ds.feedBytes([5, 6])

#         ds.feedBytes(head)
#         ds.feedBytes([11, 23])
#         print('from t2')

#     asyncio.gather(t2(), t1())

#     async def main():
#         t = await asyncio.gather(t2(), t1())
#         print(t)

#     asyncio.run(main())


# __test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : sky.li (13751020500@163.com)
# @Desc    : 协议包工具类

import asyncio
from . import config
from . import data_stream
from . import pyevent

DataStream = data_stream.DataStream
EventEmitter = pyevent.EventEmitter

_magic_head = bytes(config.MAGIC_HEAD, 'utf-8')


# 协议包打包解包
class MessagePacker:
    _emitter: EventEmitter = EventEmitter()
    _stream: DataStream = DataStream()

    def __init__(self):
        self._isPacking = False
        self._emitter: EventEmitter = EventEmitter()
        self._stream: DataStream = DataStream()

    # 得到协议包
    def pack(self, data):
        msg = [b for b in _magic_head]
        l = len(data)
        lo = (l & 0x00ff)
        hi = (l & 0xff00) >> 8
        msg.append(lo)
        msg.append(hi)
        for byte in data:
            msg.append(byte)
        return msg

    # 解协议包
    async def _unpack(self):
        if (self._isPacking):
            raise Exception('packing')
        self._isPacking = True
        while True:
            data = await self._unpackOnce()
            self._emitter.emit('unpack', data)
            if self._stream.isEmpty():
                break
        self._isPacking = False

    async def _unpackOnce(self):
        await self._stream.recvTill(_magic_head)
        lohi = await self._stream.recv(2)
        lo, hi = lohi[0], lohi[1]
        len = hi * 0xff + lo
        data = await self._stream.recv(len)
        return data

    # 当解出完整包
    def onUnpack(self, h):
        return self._emitter.on('unpack', h)

    # 送入字节
    def feedBytes(self, data):
        if (not self._isPacking):
            asyncio.ensure_future(self._unpack())
        self._stream.feedBytes(data)


# def __test():
#     import asyncio

#     async def test3():
#         print('............test MessagePacker1')
#         data = [33, 44]
#         dataRecv = None
#         s = MessagePacker()

#         def cb(d):
#             print('cb', d)
#             nonlocal dataRecv
#             dataRecv = d
#         s.onUnpack(cb)
#         s.feedBytes([1, 2])
#         await asyncio.sleep(0)
#         s.feedBytes([3, 4])
#         packed = s.pack(data)
#         print('packed', packed)
#         s.feedBytes(packed)
#         await asyncio.sleep(0.1)
#         assert (s._isPacking == False)
#         s.feedBytes([5])
#         await asyncio.sleep(0.1)
#         assert (s._isPacking == True)
#         s.feedBytes([6, 7])
#         await asyncio.sleep(0.1)
#         assert (s._isPacking == True)

#         await asyncio.sleep(0.2)
#         assert (data == dataRecv)
#         assert (s._stream._datas == [])
#         print('MessagePacker1 pass')

#     asyncio.run(test3())


# __test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : sky.li (13751020500@163.com)
# @Desc    : mlink 接口

import asyncio
import websockets
from . import config
from . import message_packer
from . import pyevent

MessagePacker = message_packer.MessagePacker
EventEmitter = pyevent.EventEmitter

# 事件泵
_emitter = EventEmitter()
_URL = "ws://localhost:"+str(config.SERVER_PORT)

# 连接后的客户端, WebSocketClientProtocol
_client = None

# 打包/解包器
_dataPacker = MessagePacker()

# 注册的数据通道
_channels = {}

# 是否关闭
_close = False

# 读入数据 Future, as Promise in nodejs
_getDataFuture = None

# 连接 Future, Future, as Promise in nodejs
_onConnectFuture = None

# 状态
_isStated = False


class _Channel:
    """数据通道类"""

    def __init__(self, chId):
        self.emitter = EventEmitter()
        self._chId = chId
        self.recvFuture = None
        self._sendloop = asyncio.get_event_loop()
        # 向server 注册该通道
        sData = [config.MSG_TAG.msg_reg, self._chId]
        packedData = _dataPacker.pack(sData)
        asyncio.run_coroutine_threadsafe(_client.send(bytes(packedData)),self._sendloop)

    def send(self, data):
        """向server 发送数据"""
        sData = [d for d in data]
        sData.insert(0, self._chId)
        sData.insert(0, config.MSG_TAG.msg_send)
        packedData = _dataPacker.pack(sData)
        asyncio.run_coroutine_threadsafe(_client.send(bytes(packedData)),self._sendloop)

    async def recv(self):
        """server 收到数据"""
        self.recvFuture = asyncio.Future()
        ret = await self.recvFuture
        self.recvFuture = None
        return ret

    def _doRecv(self, data):
        if self.recvFuture != None:
            self.recvFuture.set_result(data)

    def close(self):
        """关闭通道"""
        del _channels[self._chId]
        sData = [config.MSG_TAG.msg_unreg, self._chId]
        packedData = _dataPacker.pack(sData)
        asyncio.run_coroutine_threadsafe(_client.send(bytes(packedData)),self._sendloop)

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def doUnpack(data):
    """解包回调"""
    tag = data[0]
    if tag == config.MSG_TAG.msg_recv:
        chId = data[1]
        data = data[2:]
        ch = _channels.get(chId)
        if ch != None:
            ch._doRecv(data)
    else:
        raise Exception('wrong msg tag', tag)


def _reset():
    """重置状态"""
    global _client, _dataPacker, _channels, _close, _getDataFuture
    # 连接后的客户端, WebSocketClientProtocol
    _client = None
    # 打包/解包器
    _dataPacker = MessagePacker()
    # 注册解包回调
    _dataPacker.onUnpack(doUnpack)
    # 注册的数据通道
    _channels = {}
    # 是否关闭
    _close = False
    # 读入数据 Future, as Promise in nodejs
    _getDataFuture = None
    # 连接 Future, Future, as Promise in nodejs
    _onConnectFuture = None
    _isStated = False


async def getData():
    """协程, 从websocket数据"""
    global _client, _onConnectFuture
    try:
        async with websockets.connect(_URL) as websocket:
            print('client connected successfully to', _URL)
            _client = websocket
            _onConnectFuture.set_result(None)
            _onConnectFuture = None
            while not _close:
                try:
                    data = await websocket.recv()
                    _dataPacker.feedBytes(data)
                except Exception as ex:
                    for (_, ch) in _channels.items():
                        if(ch.recvFuture is not None):
                            ch.recvFuture.set_exception(
                                Exception('fail to recv, maybe cause by remote closing'))
                    _reset()
                    return
    except Exception as ex:
        print(ex)

    if _onConnectFuture != None:
        _onConnectFuture.set_exception(
            ('unable to connect to server'))
    _reset()


async def start():
    """ 启动服务 """
    global _onConnectFuture
    global _getDataFuture
    global _isStated
    if _isStated:
        return
    _reset()
    _isStated = True

    _onConnectFuture = asyncio.Future()
    _getDataFuture = asyncio.ensure_future(getData())
    await _onConnectFuture


async def stop():
    """ 停止服务 """
    global _close, _client
    global _isStated
    if not _isStated:
        return
    _isStated = False

    _close = True
    if _client != None:
        await _client.close()
        _client = None
    if _getDataFuture != None:
        _getDataFuture.cancel()
    _reset()


def getChannel(chId):
    """取得channel"""
    ch = _channels.get(chId)
    if ch == None:
        ch = _Channel(chId)
        _channels[chId] = ch
    return ch



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author  : sky.li (13751020500@163.com)
# @Desc    : 工程配置项

class MSG_TAG:
    msg_reg = 0  # reg(ch) // 注册一个通道
    msg_unreg = 1  # unreg(ch) // 取消注册一个通道
    msg_send = 2  # send(ch, data) // (客户端)发送数据到通道
    msg_recv = 3  # recv(ch, data) // (客户端收到)从数据通道收到数据


# 服务端 port
SERVER_PORT = 20101

# 协议头
MAGIC_HEAD = "__a4z1v1"



# -*- coding: utf-8 -*
import os
from setuptools import setup

dirs = []
modules = []
for path,dir_list,file_list in os.walk("./mlink"):
    for dir_name in dir_list:
        if dir_name.find("__")==-1 and dir_name.find("egg")==-1:
            dirs.append((path+"/"+dir_name).replace("/",".").replace("..",""))
    for file_name in file_list:
        if file_name.find(".py")!=-1 and file_name.find("__")==-1 and file_name.find(".pyc")==-1:
            modules.append((path+"."+file_name).replace(".py","").replace("/",".").replace("..",""))
# print(dirs,modules)
setup(
    name='mlink',
    version='0.0.4',
    author='makeblock',
    author_email='flashindream@gmail.com',
    url='https://makeblock.com',
    description=u'library for mlink',
    packages=dirs,
    py_modules=modules,
    install_requires=['websockets>=8.1'],
)
# setup.py
# global-include *.pyc
# global-exclude *.py